/* -*- linux-c -*- */


/*
    lm78.c - A Linux module for reading sensor data.
    Copyright (c) 1997, 1998  Alexander Larsson <alla@lysator.liu.se>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*
 *  LM78 sensor driver for LINUX
 *
 * Copyright (c) 1997 Alexander Larsson
 * 
 * provides a /proc/sensors which shows
 * some information about your system.
 * Voltages, mb temperature and fan speeds
 * are currently reported.
 *
 * This should work on all ASUS TX97 cards
 * but it's only tested on a TX97-X.
 * It should also work on any card that has
 * a LM78 circuit, but some constants in
 * lm78.h probably has to be changed.
 *
 * please report any bugs/patches to me:
 *  alla@lysator.liu.se
 *
 * Heavily patched by Frodo Looijaard <frodol@dds.nl> and Philip 
 * Edelbrock <phil@Stimpy.netroedge.com>.
 *
 * Reference:
 *   National Semiconductor
 *   LM78 Microprocessor System Hardware Monitor, September 1996.
 *   http://www.national.com/ds/LM/LM78.pdf
 *
 */

#define LM78_VERSION  "0.01"

#include <linux/config.h>

/* Added 19980620 by Frodo Looijaard (frodol@dds.nl)
   This solves undefined symbol errors for depmod/modprobe.
   Tested for Linux 2.0.34, but I do not know whether it works for other
   kernel versions. Actually, I am not even sure this is the best way to
   solve this. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/fs.h>
#include <linux/ioport.h>

#include <asm/io.h>

#include "lm78.h"

int init_module(void);
void cleanup_module(void);
static int lm78_read(char *, char **, off_t, int, int);
static int lm78_read_value(int address);
static int inodeval = 0;

struct proc_dir_entry dir =
{
	0, 7, "sensors", 
	S_IFREG | S_IRUGO, 1, 0, 0,
	0, NULL,
	&lm78_read
};

/*
 *	Initialize the module.
 */

int res;

int init_module(void)
{
  
  if (check_region(LM78_BASE, LM78_EXTENT)) 
    return -ENODEV;

  /* Verify that lm78 exists: */
  outb_p( LM78_DUMMY, LM78_ADDRESS_REG );
  res = inb_p(LM78_ADDRESS_REG);
  inb_p(LM78_DATA_REG); /* reset busy flag */

  if (res & 0x7F != LM78_DUMMY) 
    return -EIO; 

  request_region(LM78_BASE, LM78_EXTENT, "LM78");

  /* Reset all except Watchdog values and last conversion values 
     This sets fan-divs to 2, among others */
  LM78_WRITE(LM78_CONFIG, 0x80 );

  /* Start monitoring */
  LM78_WRITE(LM78_CONFIG, (LM78_READ(LM78_CONFIG) & 0xf7) | 0x01);

/*
 *	Try to register the file in the root directory of the proc
 *	filesystem.  Dynamic registration means we won't provide a fixed
 *	inode number, but rather the registration code will allocate one for
 *	us.  This is safer, since we never know when someone will add a new
 *	fixed inode number to the filesystem.  Possible returns are 0
 *	(success) and -ENOENT (failure), and since there's only one error we
 *	won't worry unduly about messages.  If successful, continue.
 */

  if (proc_register_dynamic(&proc_root, &dir)) {
    release_region(LM78_BASE, LM78_EXTENT);
    return -ENOENT;
  }
  
/*
 *      Save the inode value from the registration call, as we will need
 *      it to unregister.
 */
  inodeval = dir.low_ino;

  printk("LM78 detected and installed.\n");
  return 0;
}

/*
 *	Remove the module.  If a nonzero inode value was returned,
 *	unregister the /proc directory entry; that's all we need to do.
 */

void cleanup_module(void)
{
        release_region(LM78_BASE, LM78_EXTENT);
      
        if (inodeval)
		proc_unregister(&proc_root, inodeval);
}

/*
 *	Process a read to the file.  we just return the whole mess to the
 *	user with a new-line character on the end.
 */

static int lm78_read_value(int address) {
  while (LM78_BUSY)
    ;

  return LM78_READ(address);
}


/*
 * Ahem, gotta find out how this is supposed to work, ie. how big is 'buf' 
 */
static int lm78_read(char *buf, char **start, off_t offset, int len, int unused)
{
  int fan;
  long v;
  len=0;

#ifdef LM78_VCORE
  v = ((long) lm78_read_value(LM78_VCORE)) * 16 * NORMAL_VCORE / 100000;
  len += sprintf(buf+len, "Vcore:  %2d.%02dV\n", v/100, v%100 );
#else
  len += sprintf(buf+len, "Vcore: not supported\n");
#endif

#ifdef LM78_3_3V
  v = ((long) lm78_read_value(LM78_3_3V)) * 16 * NORMAL_3_3V / 100000;
  len += sprintf(buf+len, "+3.3V:  %2d.%02dV\n", v/100, v%100 );
#else
  len += sprintf(buf+len, "+3.3V: not supported\n");
#endif

#ifdef LM78_5V
  v = ((long) lm78_read_value(LM78_5V)) * 16 * NORMAL_5V / 100000;
  len += sprintf(buf+len, "+5.0V:  %2d.%02dV\n", v/100, v%100 );
#else
  len += sprintf(buf+len, "+5.0V: not supported\n" );
#endif

#ifdef LM78_M5V
  v = ((long) lm78_read_value(LM78_M5V)) * 16 * NORMAL_M5V / 100000;
  len += sprintf(buf+len, "-5.0V:  %2d.%02dV\n", v/100, v%100 );
#else
  len += sprintf(buf+len, "-5.0V: not supported\n" );
#endif

#ifdef LM78_12V
  v = ((long) lm78_read_value(LM78_12V)) * 16 * NORMAL_12V / 100000;
  len += sprintf(buf+len, "+12V:   %2d.%02dV\n", v/100, v%100 );
#else
  len += sprintf(buf+len, "+12V : not supported\n" );
#endif

#ifdef LM78_M12V
  v = ((long) lm78_read_value(LM78_M12V)) * 16 * NORMAL_M12V / 100000;
  len += sprintf(buf+len, "-12V:   %2d.%02dV\n", v/100, v%100 );
#else
  len += sprintf(buf+len, "-12V : not supported\n" );
#endif
  
  len += sprintf(buf+len, "Mainboard temperature: %d degrees\n", lm78_read_value(LM78_TEMPERATURE) );

  fan = LM78_READ( LM78_FAN1 );
  len += sprintf(buf+len, "Chassis fan: ");
  if (fan==255)
    len += sprintf(buf+len, "0 rpm (not present?)\n");
  else
    len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));


  fan = LM78_READ( LM78_FAN2 );
  len += sprintf(buf+len, "CPU fan: ");
  if (fan==255)
    len += sprintf(buf+len, "0 rpm (not present?)\n");
  else
    len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));


  fan = LM78_READ( LM78_FAN3 );
  len += sprintf(buf+len, "Power fan: ");
  if (fan==255)
    len += sprintf(buf+len, "0 rpm (not present?)\n");
  else
    len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));


  return len;
}




